""" This module contains custom exceptions specific to the `earthdata-varinfo`
    Python package. These exceptions are intended to allow for easier debugging
    of the expected errors that may occur during an invocation of the package,
    or any third party code that uses the package.

"""


class CustomError(Exception):
    """ Base class for exceptions in `earthdata-varinfo`. This base class
        allows for future work, such as assigning exit codes for specific
        failure modes.

    """

    def __init__(self, exception_type, message):
        self.exception_type = exception_type
        self.message = message
        super().__init__(self.message)


class DmrNamespaceError(CustomError):
    """ This exception is raised when the root element of a dmr XML document
        is not a fully qualified Dataset tag.

    """

    def __init__(self, tag):
        super().__init__('DmrNamespaceError', f'Unexpected root: {tag}')


class InvalidConfigFileFormatError(CustomError):
    """ This exception is raised when a configuration file is specified when
        creating an instance of a VarInfo or CFConfig class with a non ".json"
        extension.

    """

    def __init__(self, file_path):
        super().__init__('InvalidConfigFileFormatError',
                         f'"{file_path}" must be a JSON file.')


class InvalidExportDirectory(CustomError):
    """ This exception is raised when an output directory is specified for
        variable record export, and a file exists at that location instead.

    """

    def __init__(self, directory_path):
        super().__init__('InvalidExportDirectory',
                         f'"{directory_path}" cannot be an existing file.')


class MissingConfigurationFileError(CustomError):
    """ This exception is raised when a configuration file path is supplied to
        either a VarInfo class or the CFConfig class, but there is no file at
        the specified location.

    """

    def __init__(self, file_path):
        super().__init__('MissingConfigurationFileError',
                         f'No file in specified location: {file_path}')


class CMRQueryException(CustomError):
    ''' This exception is raised when a query to CMR fails.
    '''

    def __init__(self, cmr_exception_message):
        super().__init__('CMRQueryException',
                         'CMR query failed with the following error: '
                         f'{cmr_exception_message}')


class MissingPositionalArguments(CustomError):
    ''' This exception is raised when a function is missing a required
        positional argument.
    '''

    def __init__(self, positional_argument):
        super().__init__('MissingPositionalArguments',
                         f'Missing positional argument: {positional_argument}')


class MissingGranuleDownloadLinks(CustomError):
    ''' This exception is raised when a granule record does not contain links
        to download data.
    '''

    def __init__(self, download_link):
        super().__init__('MissingGranuleDownloadLinks',
                         f'No links for granule record: {download_link}')


class GranuleDownloadException(CustomError):
    ''' This exception is raised when the `requests.get` modules fails
        to download a variable locally.
    '''

    def __init__(self, granule_download_exception_message):
        super().__init__('GranuleDownloadException',
                         'requests module failed with the following error: '
                         f'{granule_download_exception_message}')


class DirectoryCreationException(CustomError):
    ''' This exception is raised when creating a directory fails.
    '''

    def __init__(self, directory_creation_exception_message):
        super().__init__('DirectoryCreationException',
                         'directory creation failed with the following error: '
                         f'{directory_creation_exception_message}')


class GetEdlTokenException(CustomError):
    ''' This exception is raised when `requests.post` fails to get an
        EDL token given a LaunchPad token
    '''
    def __init__(self, get_edl_token_exception_message):
        super().__init__('GetEdlTokenException',
                         'requests module failed with the following error: '
                         f'{get_edl_token_exception_message}')



""" This module contains a class designed to read and present information from
    a JSON configuration file. This configuration file provides supplemental
    and overriding information for attributes provided by granules on a per-
    collection basis. This information is primarily intended to augment the
    CF-Convention attributes for a dataset, but can also be used to alter non
    CF-Convention metadata within a granule.

    Information within the configuration file is split into blocks that have
    an Applicability_Group. This section should define a mission, collection
    short name and (optionally) a regular expression compatible string for
    relevant variable paths. These applicability groups can be nested, and so
    the mission and short name can be inherited from the parent group.

    The configuration file also specifies variables that are incorrectly
    considered as science variables by the VarInfo class, due to them having
    references in the coordinates attribute. Further, there are required
    variables that have to be included in the output of any variable subset
    request for a specific collection.

"""
from os.path import exists
from typing import Dict, Optional
import json
import re

from varinfo.exceptions import (InvalidConfigFileFormatError,
                                MissingConfigurationFileError)


class CFConfig:
    """ This class should read the main configuration file,
        see e.g. sample_config.json, which defines overriding values and
        supplements for the attributes stored in fields such as
        ancillary_variables, or dimensions.

        Given a mission and collection short name, upon instantiation, the
        object should only retain information relevant to that specific
        collection.

    """
    def __init__(self, mission: str, collection_short_name: str,
                 config_file: Optional[str] = None):
        """ Set supplied class attributes. Then read the designated
            configuration file to obtain mission and short name specific
            attributes.

        """
        self.config_file = config_file
        self.mission = mission
        self.short_name = collection_short_name

        self._cf_overrides = {}
        self._cf_supplements = {}
        self.excluded_science_variables = set()
        self.required_variables = set()
        self.global_supplements = {}
        self.global_overrides = {}

        if self.mission is not None:
            self._read_config_file()

    def _read_config_file(self):
        """ Open the main configuration JSON file and extract only those parts
            of it pertaining to the mission and collection specified upon
            instantiating the class.

        """
        if self.config_file is not None and not exists(self.config_file):
            raise MissingConfigurationFileError(self.config_file)
        elif (
            self.config_file is not None and self.config_file.endswith('.json')
        ):
            with open(self.config_file, 'r', encoding='utf-8') as file_handler:
                config = json.load(file_handler)
        elif self.config_file is not None:
            raise InvalidConfigFileFormatError(self.config_file)
        else:
            config = {}

        self.excluded_science_variables = {
            pattern
            for item in config.get('Excluded_Science_Variables', [])
            if self._is_applicable(item['Applicability'].get('Mission'),
                                   item['Applicability'].get('ShortNamePath'))
            for pattern in item['Variable_Pattern']
        }

        self.required_variables = {
            pattern
            for item in config.get('Required_Fields', [])
            if self._is_applicable(item['Applicability'].get('Mission'),
                                   item['Applicability'].get('ShortNamePath'))
            for pattern in item['Variable_Pattern']
        }

        self.global_supplements = {
            attribute['Name']: attribute['Value']
            for item in config.get('CF_Supplements', [])
            if self._is_applicable(item['Applicability'].get('Mission'),
                                   item['Applicability'].get('ShortNamePath'))
            for attribute in item.get('Global_Attributes', [])
        }

        self.global_overrides = {
            attribute['Name']: attribute['Value']
            for item in config.get('CF_Overrides', [])
            if self._is_applicable(item['Applicability'].get('Mission'),
                                   item['Applicability'].get('ShortNamePath'))
            for attribute in item.get('Global_Attributes', [])
        }

        for override in config.get('CF_Overrides', []):
            self._process_cf_item(override, self._cf_overrides)

        for supplement in config.get('CF_Supplements', []):
            self._process_cf_item(supplement, self._cf_supplements)

    def _is_applicable(self, mission: str, short_name: str = None) -> bool:
        """ Given a mission, and optionally also a collection short name, of an
            applicability within the configuration file, check for a match
            against the mission and short name specified when instantiating the
            class object.

        """
        mission_matches = re.match(mission, self.mission) is not None

        short_name_matches = (
            short_name is None
            or re.match(short_name, self.short_name) is not None
        )

        return mission_matches and short_name_matches

    def _process_cf_item(self,
                         cf_item: Dict,
                         results: Dict[str, Dict],
                         input_mission: str = None,
                         input_short_name: str = None):
        """ Process a single block in the CF overrides or CF supplements region
            of the configuration file. First check that the applicability
            matches the mission and short name for the class. Next, check
            for a variable pattern. This is indicative of there being
            overriding or supplemental attributes in this list item.
            Assign any information to the results dictionary, with a key of
            that variable pattern. Lastly, check for any nested references,
            which are child blocks to be processed in the same way. The mission
            and short name from this block are passed to all children, as they
            may not both be defined, due to assumed inheritance.

        """
        mission = cf_item['Applicability'].get('Mission') or input_mission
        short_name = cf_item['Applicability'].get('ShortNamePath') or input_short_name

        if mission is not None and self._is_applicable(mission, short_name):
            # Some outer Applicability items have attributes, but no
            # variable path - the assumption here is that the applicability is
            # to all variables (see ICESat2 dimensions override, SPL4.* and
            # SPL3FTA grid_mapping overrides)
            pattern = cf_item['Applicability'].get('Variable_Pattern', '.*')

            if 'Attributes' in cf_item:
                results[pattern] = self._create_attributes_object(cf_item)

            cf_references = cf_item.get('Applicability_Group', [])

            for cf_reference in cf_references:
                self._process_cf_item(cf_reference, results, mission,
                                      short_name)

    @staticmethod
    def _create_attributes_object(cf_item: Dict) -> Dict[str, str]:
        """ Construct a dictionary object containing all contained attributes,
            which are specified as list items with Name and Value keys.

        """
        return {attribute['Name']: attribute['Value']
                for attribute in cf_item.get('Attributes', {})}

    def get_cf_attributes(self, variable: str = None) -> Dict[str, Dict[str, str]]:
        """ Return the CF overrides and supplements that match a given
            variable. If a variable is not specified, then return all overrides
            and supplements. If there are no overrides or supplements, then
            empty dictionaries will be returned instead.

        """
        if variable is not None:
            cf_overrides = self._get_matching_attributes(self._cf_overrides,
                                                         variable)
            cf_supplements = self._get_matching_attributes(self._cf_supplements,
                                                           variable)
        else:
            cf_overrides = self._cf_overrides
            cf_supplements = self._cf_supplements

        return {'cf_overrides': cf_overrides, 'cf_supplements': cf_supplements}

    @staticmethod
    def _get_matching_attributes(cf_references: Dict[str, Dict[str, str]],
                                 variable: str) -> Dict[str, str]:
        """ Iterate through either the self._cf_supplements or
            self._cf_overrides and extract a dictionary that combines all
            applicable attributes that apply to the specified variable. If
            there are conflicting values for the same attribute, only the last
            value will be returned for that attribute.

        """
        references = {}
        for pattern, attributes in cf_references.items():
            if re.match(pattern, variable) is not None:
                references.update(attributes)

        return references



""" Functionality to convert a Variable instance, from either a DMR or NetCDF-4
    input, to a UMM-Var record.

    To use:

    ```
    from varinfo import VarInfoFromNetCDF4
    from varinfo.umm_var import export_all_umm_var, get_all_umm_var

    varinfo_gpm = VarInfoFromNetCDF4('/path/to/GPM_3IMERGHH/granule.nc4',
                                     short_name='GPM_3IMERGHH')

    umm_var_records = get_all_umm_var(varinfo_gpm)

    # To export to disk:
    export_all_umm_var(umm_var_records.values(), output_dir='path/to/directory')
    ```

    To validate generated records:

    ```
    import json
    from jsonschema import validate

    with open('tests/unit/data/umm_var_json_schema_1.8.2.json') as file_handler:
        umm_var_schema = json.load(file_handler)

    for record in umm_var_records:
        validate(schema=umm_var_schema, instance=umm_var_schema)
    ```

"""
from os import makedirs
from os.path import isfile, join as join_path
from typing import Any, Dict, List, Optional, Union
import json
import requests

from cmr import CMR_UAT
from numpy import floating as np_floating, integer as np_integer

from varinfo.cmr_search import CmrEnvType
from varinfo.exceptions import InvalidExportDirectory
from varinfo.var_info import VarInfoBase
from varinfo.variable import VariableBase


UMM_URL = 'https://cdn.earthdata.nasa.gov/umm'
UMM_VAR_VERSION = '1.8.2'

UMM_VAR_DTYPES = ['byte', 'float', 'float32', 'float64', 'double', 'ubyte',
                  'ushort', 'uint', 'uchar', 'string', 'char8', 'uchar8',
                  'short', 'long', 'int', 'int8', 'int16', 'int32', 'int64',
                  'uint8', 'uint16', 'uint32', 'uint64', 'OTHER']


def get_all_umm_var(var_info: VarInfoBase) -> Dict[str, Dict]:
    """ Iterate through all variables detected from the source granule and
        return a list of UMM-Var records for those variables.

    """
    return {variable_name: get_umm_var(var_info, variable)
            for variable_name, variable in var_info.variables.items()}


def get_umm_var(var_info: VarInfoBase, variable: VariableBase) -> Dict:
    """ Map the contents of a Variable instance to a UMM-Var record.
        Initial attempts will be made to extract all possibly information,
        with the return value being scrubbed of all UMM-Var attributes with
        None values.

    """
    variable_name = variable.full_name_path.lstrip('/')
    umm_var_record = {
        'Name': variable_name,
        'LongName': get_first_matched_attribute(variable, ['long_name'],
                                                variable_name),
        'StandardName': get_first_matched_attribute(variable,
                                                    ['standard_name']),
        'Definition': get_first_matched_attribute(
            variable,
            ['description', 'Description', 'definition', 'Definition',
             'title', 'Title'],
            variable_name
        ),
        'DataType': get_umm_var_dtype(variable.data_type),
        'Dimensions': get_dimensions(var_info, variable),
        'Units': get_first_matched_attribute(variable, ['units', 'Units']),
        'FillValues': get_fill_values(variable),
        'Scale': get_first_matched_attribute(
            variable, ['scale_factor', 'scale', 'Scale']
        ),
        'Offset': get_first_matched_attribute(
            variable, ['add_offset', 'offset', 'Offset']
        ),
        'ValidRanges': get_valid_ranges(variable),
        'VariableType': get_variable_type(var_info, variable),
        'MetadataSpecification': get_metadata_specification()
    }

    return {umm_var_attribute: umm_var_value
            for umm_var_attribute, umm_var_value
            in umm_var_record.items()
            if umm_var_value is not None}


def export_all_umm_var_to_json(umm_var_records: List, output_dir: str = '.'):
    """ Iterate through a list of UMM-Var JSON records and save them all to
        files (one file per record).

        The output file name will be the full path of each variable will be the
        full path of the variable, with any slashes replaced with underscores.

    """
    for umm_var_record in umm_var_records:
        export_umm_var_to_json(umm_var_record, output_dir)


def export_umm_var_to_json(umm_var_record: Dict, output_dir: str = '.'):
    """ Export a single UMM-Var JSON object to a JSON output file in a
        specified directory (the default is the directory in which the function
        is called). If the specified directory does not exist, it will be
        created.

        The output file name will be the full path of the variable, with any
        slashes replaced with underscores.

    """
    if isfile(output_dir):
        raise InvalidExportDirectory(output_dir)
    else:
        # `exists_ok=True` makes this a no-op for existing directories:
        makedirs(output_dir, exist_ok=True)

    output_file_path = join_path(
        output_dir,
        f'{umm_var_record["Name"].replace("/", "_")}.json'
    )

    with open(output_file_path, 'w', encoding='utf-8') as file_handler:
        json.dump(umm_var_record, file_handler, indent=2)


def get_first_matched_attribute(variable: VariableBase,
                                attribute_names: List[str],
                                default_value: Any = None) -> Any:
    """ Check a list of metadata attributes and return the value of the first
        one that is present in the Variable. If none of the attributes are
        in the variable metadata, return the supplied default value.

    """
    return next(
        (get_json_serializable_value(variable.attributes.get(attribute_name))
         for attribute_name in attribute_names
         if attribute_name in variable.attributes),
        default_value
    )


def get_dimensions(var_info: VarInfoBase,
                   variable: VariableBase) -> Optional[List]:
    """ Return a list of all dimensions for the variable, """
    dimensions = [get_dimension_information(var_info, variable, dimension_name)
                  for dimension_name in variable.dimensions]

    if len(dimensions) == 0:
        dimensions = None

    return dimensions


def get_dimension_information(var_info: VarInfoBase, variable: VariableBase,
                              dimension_name: str) -> Dict:
    """ Retrieve a DimensionType object for the given Variable dimension. This
        function is only called for named dimensions listed in the `dimensions`
        attribute of another variable, and so should exist as at least a
        dimension within the source file.

        The dimension types in earthdata-varinfo are currently limited to
        horizontal spatial dimensions (e.g., lat, lon, projected x or projected
        y), or temporal dimensions based on the available heuristics in the
        VariableBase class.

        The dimension Name property is the full path to the variable, omitting
        any leading slashes. (E.g., a variable located at "/group/variable"
        within a file will extract a name of "group/variable")

        Currently unsupported dimension types (will have type: 'OTHER'):

        * Vertical dimensions (pressure, height, depth).
        * Projected horizontal spatial dimension (no UMM-Var type for this).
        * Swath dimensions (along or across track).

        Also note: VariableFromDmr instances do not currently extract variable
        shapes.

    """
    dimension_variable = var_info.get_variable(dimension_name)

    if dimension_variable is not None:
        dimension_name = dimension_variable.full_name_path
        if dimension_variable.is_latitude():
            dimension_type = 'LATITUDE_DIMENSION'
        elif dimension_variable.is_longitude():
            dimension_type = 'LONGITUDE_DIMENSION'
        elif dimension_variable.is_temporal():
            dimension_type = 'TIME_DIMENSION'
        else:
            dimension_type = 'OTHER'

    else:
        # Dimension without variable, potentially used only to denote array
        # size in a specific, non-physical dimension (e.g., nv, latv, lonv).
        dimension_type = 'OTHER'

    return {'Name': dimension_name.lstrip('/'),
            'Size': get_dimension_size(var_info, variable, dimension_name),
            'Type': dimension_type}


def get_dimension_size(var_info: VarInfoBase, variable_with_dim: VariableBase,
                       dimension_name: str) -> Union[str, int]:
    """ Extract the size of a specific dimension for a variable. This
        function will attempt to retrieve the dimension size from the following
        locations (in the order given):

        * The shape of the variable with the dimension itself, if present.
        * The shape of the 1-D dimension variable corresponding to the
          dimension for which the size is requested.
        * The variable will be checked for known names, as described in the
          CF-Conventions, to assign a length (relating to bounds variables).
        * If no other condition is met, the size of 'Varies' is returned.

        Note, this function is only called via an iteration through the
        `variable_with_dim.dimensions` list, and so `dimension_name` should
        always be present in that list.

    """
    dimension_variable = var_info.get_variable(dimension_name)

    if variable_with_dim.shape is not None:
        dimension_index = variable_with_dim.dimensions.index(dimension_name)
        dimension_size = variable_with_dim.shape[dimension_index]
    elif dimension_variable is not None:
        if isinstance(dimension_variable.shape, int):
            dimension_size = dimension_variable.shape
        elif (
            isinstance(dimension_variable.shape, tuple)
            and len(dimension_variable.shape) > 0
        ):
            dimension_size = dimension_variable.shape[0]
        else:
            # UMM-Var requires a size for each dimension, which must be either
            # an integer or 'Varies':
            dimension_size = 'Varies'
    elif dimension_name.endswith(('latv', 'lonv', 'nv')):
        # Bound variable dimensions (2-elements, minimum and maximum):
        dimension_size = 2
    else:
        # Other dimension without variable, potentially used only to denote
        # array size in a specific, non-physical dimension
        dimension_size = 'Varies'

    return dimension_size


def get_valid_ranges(variable: VariableBase) -> Optional[List[Dict]]:
    """ Return a dictionary containing the valid minimum and/or valid maximum
        values from the variable metadata. If valid_min, valid_max or
        valid_range are not set, None is returned.

    """
    valid_range = {
        'Min': get_json_serializable_value(variable.get_valid_min()),
        'Max': get_json_serializable_value(variable.get_valid_max())
    }

    # Remove keys with None values:
    valid_range = {range_key: range_value
                   for range_key, range_value in valid_range.items()
                   if range_value is not None}

    if len(list(valid_range.keys())) > 0:
        valid_range = [valid_range]
    else:
        valid_range = None

    return valid_range


def get_fill_values(variable: VariableBase) -> Optional[List]:
    """ Return a List containing elements of the UMM-Var FillValueType, if
        there is a fill value contained in the variable metadata. Otherwise
        return None.

        An initial simplification is that all fill values are of type
        `SCIENCE_FILLVALUE`.

    """
    fill_value = get_first_matched_attribute(variable, ['_FillValue'])
    if fill_value is not None:
        fill_values = [{
            'Value': get_json_serializable_value(fill_value),
            'Type': 'SCIENCE_FILLVALUE',
            'Description': 'Extracted from _FillValue metadata attribute'
        }]
    else:
        fill_values = None

    return fill_values


def get_umm_var_dtype(variable_data_type: str) -> str:
    """ Map the variable data type to a string in the UMM-Var DataTypeEnum. """
    if variable_data_type in UMM_VAR_DTYPES:
        umm_var_type = variable_data_type
    else:
        umm_var_type = 'OTHER'

    return umm_var_type


def get_metadata_specification() -> Dict:
    """ Return standard object for the UMM-Var specification, including the
        URL, Name and Version.

    """
    return {'URL': f'{UMM_URL}/variable/v{UMM_VAR_VERSION}',
            'Name': 'UMM-Var',
            'Version': UMM_VAR_VERSION}


def get_json_serializable_value(input_value: Any) -> Any:
    """ Ensure the value is JSON serializable, as some numpy float and integer
        types are not.

    """
    if isinstance(input_value, np_integer):
        output_value = int(input_value)
    elif isinstance(input_value, np_floating):
        output_value = float(input_value)
    else:
        output_value = input_value

    return output_value


def generate_variable_native_id(collection_concept_id: str,
                                umm_var_record: Dict) -> str:
    """ A helper function to create a CMR native ID given the collection
        concept ID and the variable UMM-Var JSON. This native ID must be unique
        within the entire provider. The initial implementation will be to
        concatenate the collection concept ID and the name of the variable
        while removing slashes that CMR will interpret as part of the URL path.

        Note - the `Name` attribute of the generated UMM-Var record is the full
        path of the variable.

    """
    return '-'.join([collection_concept_id,
                     umm_var_record['Name'].replace('/', '_').lstrip('_')])


def get_variable_type(var_info: VarInfoBase, variable: VariableBase) -> str:
    """ Check if a variable is a science variable and
        map it to a 'SCIENCE_VARIABLE' if it is.
    """
    if var_info.is_science_variable(variable):
        variable_type = 'SCIENCE_VARIABLE'
    else:
        variable_type = None

    return variable_type


def publish_umm_var(collection_id: str,
                    umm_var_dict: Dict,
                    auth_header: str,
                    cmr_env: CmrEnvType = CMR_UAT) -> str:
    """" Publish a single UMM-Var entry to CMR given:
        * collection_id: a collection's concept_id
        * umm_var_dict: a dictionary of a single UMM-Var entry for a collection
        * auth_header: Authorization HTTP header, containing a LaunchPad
          token: 'Authorization: <token>'
        * cmr_env: CMR environments (OPS, UAT, and SIT) default is CMR_UAT

        For a successful requests all of these fields must be entered

    """
    # Required UMM-Var headers for ingesting variable entries
    headers_umm_var = {
        'Content-type': 'application/vnd.nasa.cmr.umm+json;version='
        + f'{umm_var_dict["MetadataSpecification"]["Version"]}',
        'Authorization': auth_header,
        'Accept': 'application/json'
    }
    variable_native_id = generate_variable_native_id(collection_id,
                                                     umm_var_dict)

    url_endpoint = (cmr_env.replace('search', 'ingest') + 'collections/'
                    f'{collection_id}/variables/{variable_native_id}')

    response = requests.put(url_endpoint,
                            json=umm_var_dict,
                            headers=headers_umm_var,
                            timeout=10)

    if response.ok:
        # A successful request returns the variable concept-id
        # e.g., 'V1259791517-EEDTEST'
        return response.json()['concept-id']
    else:
        # A failed request returns the response containing a list of error
        # message, e.g., {'errors': ['#: required key [LongName] not found']}.
        # This will be converted into a single string that can be returned to
        # the end-user. Multiple errors will be combined into a single string,
        # e.g.: '#: CMR error 1\n  #: CMR error 2'
        return '\n  '.join(response.json()['errors'])


def publish_all_umm_var(collection_id: str,
                        all_umm_var_dict: Dict,
                        auth_header: str,
                        cmr_env: CmrEnvType = CMR_UAT) -> Dict[str, str]:
    """ Publish all UMM-Var entries associated with a collection to CMR given:
        * collection_id: a collection's concept_id
        * all_umm_var_dict: a nested dictionary containing
            dictionaries of all UMM-Var entries for a collection
        * auth_header: Authorization HTTP header, containing a LaunchPad
          token: 'Authorization: <token>'
        * cmr_env: CMR environments (OPS, UAT, and SIT)
    For a successful requests all of these fields must be entered
    """
    return {
        var_name: publish_umm_var(collection_id, umm_var, auth_header, cmr_env)
        for var_name, umm_var in all_umm_var_dict.items()
    }



""" Module containing convenience exports for the earthdata-varinfo library."""
from .cf_config import CFConfig  # noqa
from .var_info import VarInfoFromDmr, VarInfoFromNetCDF4  # noqa
from .variable import VariableFromDmr, VariableFromNetCDF4  # noqa



from pathlib import Path
from unittest import TestCase

from pycodestyle import StyleGuide


class TestCodeFormat(TestCase):
    """ This test class should ensure all earthdata-varinfo Python code adheres
        to standard Python code styling.

        Ignored errors and warnings:

        * E501: Line length, which defaults to 80 characters. This is a
                preferred feature of the code, but not always easily achieved.
        * W503: Break before binary operator. Have to ignore one of W503 or
                W504 to allow for breaking of some long lines. PEP8 suggests
                breaking the line before a binary operator is more "Pythonic".

    """
    @classmethod
    def setUpClass(cls):
        cls.python_files = Path('varinfo').rglob('*.py')

    def test_pycodestyle_adherence(self):
        """ Ensure all code in the `varinfo` directory adheres to PEP8 defined
            standards.

        """
        style_guide = StyleGuide(ignore=['E501', 'W503'])
        results = style_guide.check_files(self.python_files)
        self.assertEqual(results.total_errors, 0, 'Found code style issues.')



''' This module generates UMM-Var records for a specified collection, and
    optionally will publish those records to CMR.
    `generation_collection_umm_var` will:

    * Performing a CMR granule search with collection-based query parameters.
    * Downloading the granule locally.
    * Parse the variable metadata from the granule using VarInfoFromNetCDF4.
    * Generate UMM-Var JSON as a dictionary for each identified variable.
    * (Optional) Publish each UMM-Var entry to the selected CMR environment.

'''
from tempfile import TemporaryDirectory
from typing import Dict, List, Union
import re

from cmr import CMR_UAT

from varinfo import VarInfoFromNetCDF4
from varinfo.cmr_search import (CmrEnvType, download_granule, get_granule_link,
                                get_granules, get_edl_token_header)
from varinfo.umm_var import get_all_umm_var, publish_all_umm_var


# Custom return type: either a list of UMM-Var JSON (a list of dictionaries),
# or a list of strings (either concept IDs or error strings).
UmmVarReturnType = List[Union[Dict, str]]


def generate_collection_umm_var(collection_concept_id: str,
                                auth_header: str,
                                cmr_env: CmrEnvType = CMR_UAT,
                                publish: bool = False) -> UmmVarReturnType:
    """ Run all the of the functions for downloading and publishing
        a UMM-Var entry to CMR given:

        * collection_concept_id: Concept ID for collection that variables have
          been generated for.
        * cmr_env: URLs for CMR environments (OPS, UAT, and SIT)
        * auth_header: Authorization HTTP header, containing a LaunchPad
          token: 'Authorization: <token>'
        * publish: Optional argument determining whether to publish the
          generated UMM-Var records to the indicated CMR instance. Defaults to
          False.

        Note - if attempting to publish to CMR, a LaunchPad token must be used.

    """
    # Get an EDL token and its header given a LaunchPad token
    auth_header_edl_token = get_edl_token_header(auth_header, cmr_env)

    granule_response = get_granules(collection_concept_id, cmr_env=cmr_env,
                                    auth_header=auth_header_edl_token)

    # Get the data download URL for the most recent granule (NetCDF-4 file)
    granule_link = get_granule_link(granule_response)

    with TemporaryDirectory() as temp_dir:
        # Download file to runtime environment
        local_granule = download_granule(granule_link,
                                         auth_header_edl_token,
                                         out_directory=temp_dir)

        # Parse the granule with VarInfo to map all variables and relations:
        var_info = VarInfoFromNetCDF4(local_granule)

        # Generate all the UMM-Var records:
        all_umm_var_records = get_all_umm_var(var_info)

    if publish:
        # Publish to CMR and construct an output object that is a list of
        # strings. These strings will be either variable concept IDs or
        # error messages returned from CMR.
        publication_response = publish_all_umm_var(collection_concept_id,
                                                   all_umm_var_records,
                                                   auth_header, cmr_env)

        # Produce a list indicating publication information for all variables.
        # Variables that were successfully published will have a list element
        # providing their variable concept ID. Any variables that were
        # not successfully published will instead have an element containing
        # the variable name and the CMR error (e.g., 'variable: CMR error...').
        return_value = [variable_response
                        if is_variable_concept_id(variable_response)
                        else ': '.join([variable_name, variable_response])
                        for variable_name, variable_response
                        in publication_response.items()]
    else:
        # If not publishing, return the full UMM-Var JSON records
        return_value = list(all_umm_var_records.values())

    return return_value


def is_variable_concept_id(possible_concept_id: str) -> bool:
    """ A helper function to identify if a given string conforms to the
        expected structure of a variable concept ID, e.g., 'V1234567890-PROV'.

    """
    return bool(re.match(r'^V\d{10}-\w+$', possible_concept_id))
